import numpy as np
import scipy.stats as stats
import scipy.optimize as optimize

from . import utils as cp_utils
from . import evaluate_dmgs_equation as cp_dmgs
from . import genpareto_derivs as cp_gpd_c
from . import genpareto_libs as cp_gpd_b


def ppf(x, p=None, kloc=0, ics=None, fdalpha=0.01, customprior=0,
               minxi=-1, maxxi=2.0, means=False, waicscores=False, extramodels=False,
               pdf=False, dmgs=True, rust=False, nrust=100000, debug=False):
    """
    Passed data from the Generalized Pareto Distribution with known location parameter, returns quantiles and other results.
    
    The GP distribution has exceedcance distribution function
    S(x;μ,σ,ξ) = [1+ξ(x-μ)/σ]^(-1/ξ) if ξ ≠ 0
                 exp(-(x-μ)/σ) if ξ = 0
    
    where x is the random variable and μ,σ>0,ξ are the parameters.
    
    The calibrating prior we use is given by
    π(μ,σ,ξ) ∝ 1/σ
    as given in Jewson et al. (2025).
    
    The code will stop with an error if the input data gives a maximum likelihood
    value for the shape parameter that lies outside the range (minxi,maxxi),
    since outside this range there may be numerical problems.
    Such values seldom occur in real observed data for maxima.

    Parameters
    ----------
    x : array_like
        Input data array.
    p : array_like, optional
        Probabilities for quantile calculation (default is np.arange(0.1, 1.0, 0.1)).
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0]).
    fdalpha : float, optional
        Finite difference step for PDF estimation (default is 0.01).
    means : bool, optional
        Whether to compute means for extra models (default is False).
    waicscores : bool, optional
        Whether to compute WAIC scores (default is False).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    pdf : bool, optional
        Whether to compute PDFs (default is False).
    customprior : float, optional
        Custom prior value for shape parameter (default is 0).
    dmgs : bool, optional
        Whether to use the DMGS method (default is True).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, quantiles, PDFs, means, WAIC scores, and other results.
    """
    
    # 1 intro
    if p is None:
        p = np.arange(0.1, 1.0, 0.1)
    if ics is None:
        ics = np.array([0, 0])
    
    x = cp_utils.to_array(x)
    p = cp_utils.to_array(p)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert np.all(np.isfinite(p)) and not np.any(np.isnan(p)), "p must be finite and not NA"
    assert np.all(p > 0) and np.all(p < 1), "p must be between 0 and 1"
    assert len(ics) == 2, "ics must have length 2"
    assert not np.any(x < 0), "x must be non-negative"
    
    alpha = 1 - p
    nx = len(x)
    nalpha = len(alpha)
    
    if pdf:
        dalpha = np.minimum(fdalpha * alpha, fdalpha * (1 - alpha))
        alpham = alpha - dalpha
        alphap = alpha + dalpha
    
    # 2 ml param estimate
    ics = cp_gpd_b.gpd_k1_setics(x, ics)
    opt1 = optimize.minimize(lambda params: -cp_gpd_b.gpd_k1_loglik(params, x, kloc), 
                           ics, method='BFGS')
    v1hat = opt1.x[0]   # sigma
    v2hat = opt1.x[1]   # xi
    ml_params = np.array([v1hat, v2hat])
    
    # gpd_k1_checkmle(ml_params,kloc,minxi,maxxi)
    if debug:
        print(f"    v1hat,v2hat={v1hat},{v2hat}")
    
    if (abs(v2hat) >= 1) or (v2hat > 30):
        revert2ml = True
    else:
        revert2ml = False
    
    # 3 aic
    ml_value = opt1.fun
    maic = cp_utils.make_maic(ml_value, nparams=2)
    
    # 4 calc ml quantiles and densities (vectorized over alpha)
    ml_quantiles = stats.genpareto.ppf(1 - alpha, v2hat, loc=kloc, scale=v1hat)
    
    if v2hat < 0:
        ml_max = kloc - v1hat / v2hat
    else:
        ml_max = np.inf
    
    fhat = stats.genpareto.pdf(ml_quantiles, v2hat, loc=kloc, scale=v1hat)
    
    # dmgs
    standard_errors = "dmgs not selected"
    cp_quantiles = "dmgs not selected"
    rh_flat_quantiles = "dmgs not selected"
    lp_quantiles = "dmgs not selected"
    lp2_quantiles = "dmgs not selected"
    dpi_quantiles = "dmgs not selected"
    ru_quantiles = "dmgs not selected"
    ml_pdf = "dmgs not selected"
    cp_pdf = "dmgs not selected"
    rh_flat_pdf = "dmgs not selected"
    waic1 = "dmgs not selected"
    waic2 = "dmgs not selected"
    ml_mean = "dmgs not selected"
    rh_flat_mean = "dmgs not selected"
    cp_mean = "dmgs not selected"
    cp_method = "dmgs not selected"
    
    if dmgs and not revert2ml:
        # 5 alpha pdf stuff
        if pdf:
            xi = v2hat  # xi = c here
            ml_quantilesm = stats.genpareto.ppf(1 - alpham, xi, loc=kloc, scale=v1hat)
            ml_quantilesp = stats.genpareto.ppf(1 - alphap, xi, loc=kloc, scale=v1hat)
            fhatm = stats.genpareto.pdf(ml_quantilesm, xi, loc=kloc, scale=v1hat)
            fhatp = stats.genpareto.pdf(ml_quantilesp, xi, loc=kloc, scale=v1hat)
        
        if debug:
            print(f"ml_quantiles={ml_quantiles}")
        
        # 7 ldd (two versions)
        if debug:
            print("calculate ldd")
        
        ldd = cp_gpd_c.gpd_k1_ldda(x, v1hat, v2hat, kloc)
        lddi = np.linalg.inv(ldd)
        standard_errors = cp_utils.make_se(nx, lddi)
        
        if debug:
            print(f"ldd={ldd}")
        
        # 8 lddd (two versions)
        if debug:
            print("calculate lddd")
        
        lddd = cp_gpd_c.gpd_k1_lddda(x, v1hat, v2hat, kloc)
        
        # 9 mu1 (two versions)
        mu1 = cp_gpd_c.gpd_k1_mu1fa(alpha, v1hat, v2hat, kloc)
        
        if pdf:
            mu1m = cp_gpd_c.gpd_k1_mu1fa(alpham, v1hat, v2hat, kloc)
            mu1p = cp_gpd_c.gpd_k1_mu1fa(alphap, v1hat, v2hat, kloc)
        
        # 10 mu2 (two versions)
        mu2 = cp_gpd_c.gpd_k1_mu2fa(alpha, v1hat, v2hat, kloc)
        
        if pdf:
            mu2m = cp_gpd_c.gpd_k1_mu2fa(alpham, v1hat, v2hat, kloc)
            mu2p = cp_gpd_c.gpd_k1_mu2fa(alphap, v1hat, v2hat, kloc)
        
        # 13 model 4: rh_Flat with flat prior on shape (needs to use 3d version of Bayesian code)
        lambdad_rh_flat = np.array([-1/v1hat, 0])
        dq = cp_dmgs.dmgs(lddi, lddd, mu1, lambdad_rh_flat, mu2, dim=2)
        rh_flat_quantiles = ml_quantiles + dq / (nx * fhat)
        
        if pdf:
            dqm = cp_dmgs.dmgs(lddi, lddd, mu1m, lambdad_rh_flat, mu2m, dim=2)
            dqp = cp_dmgs.dmgs(lddi, lddd, mu1p, lambdad_rh_flat, mu2p, dim=2)
            quantilesm = ml_quantilesm + dqm / (nx * fhatm)
            quantilesp = ml_quantilesp + dqp / (nx * fhatp)
            ml_pdf = fhat
            rh_flat_pdf = -(alphap - alpham) / (quantilesp - quantilesm)
        else:
            ml_pdf = fhat
            rh_flat_pdf = "pdf not selected"
        
        # 15 model 6: Laplace's method
        if extramodels or means:
            lambdad_lp = np.array([0, 0])
            lddd_lp = np.zeros((2, 2, 2))
            dq = cp_dmgs.dmgs(lddi, lddd_lp, mu1, lambdad_lp, mu2, dim=2)
            lp_quantiles = ml_quantiles + dq / (nx * fhat)
        else:
            lp_quantiles = "extramodels not selected"
        
        # 16 model 7: Laplace's method, but with 1/sigma prior
        if extramodels or means:
            lambdad_lp2 = np.array([-1/v1hat, 0])
            lddd_lp2 = np.zeros((2, 2, 2))
            dq = cp_dmgs.dmgs(lddi, lddd_lp2, mu1, lambdad_lp2, mu2, dim=2)
            lp2_quantiles = ml_quantiles + dq / (nx * fhat)
        else:
            lp2_quantiles = "extramodels not selected"
        
        # 17 model 8: user defined xi gradient of log prior
        if extramodels or means:
            lambdad_dpi = np.array([-1/v1hat, customprior])
            dq = cp_dmgs.dmgs(lddi, lddd, mu1, lambdad_dpi, mu2, dim=2)
            dpi_quantiles = ml_quantiles + dq / (nx * fhat)
        else:
            dpi_quantiles = "extramodels not selected"
        
        # 18 means
        means_result = cp_gpd_b.gpd_k1_means(means, ml_params, kloc=kloc)
        
        ml_mean = means_result['ml_mean']
        rh_flat_mean = means_result['rh_flat_mean']
        
        # 19 waicscores
        waic = cp_gpd_b.gpd_k1_waic(waicscores, x, v1hat, v2hat, kloc, lddi, lddd,
                                  lambdad_rh_flat)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
        
        # 21 rust
        ru_quantiles = "rust not selected"
        if rust:
            rustsim = rvs(nrust, x, kloc=kloc, rust=True, mlcp=False)
            ru_quantiles = cp_utils.makeq(rustsim['ru_deviates'], p)
    
    else:
        rh_flat_quantiles = ml_quantiles
        ru_quantiles = ml_quantiles
        lp_quantiles = ml_quantiles
        lp2_quantiles = ml_quantiles
        dpi_quantiles = ml_quantiles
        rh_flat_pdf = ml_pdf
        rh_flat_mean = ml_mean
    # end of if(dmgs)
    
    return {
        'ml_params': ml_params,
        'ml_value': ml_value,
        # 'ldd': ldd,
        # 'lddi': lddi,
        'standard_errors': standard_errors,
        'ml_quantiles': ml_quantiles,
        'ml_max': ml_max,
        'revert2ml': revert2ml,
        'cp_quantiles': rh_flat_quantiles,
        'ru_quantiles': ru_quantiles,
        'lp_quantiles': lp_quantiles,
        'lp2_quantiles': lp2_quantiles,
        'dpi_quantiles': dpi_quantiles,
        'ml_pdf': ml_pdf,
        'cp_pdf': rh_flat_pdf,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_mean': ml_mean,
        'cp_mean': rh_flat_mean,
        'cp_method': cp_utils.crhpflat_dmgs_cpmethod()
    }


def rvs(n, x, kloc=0, ics=None, minxi=-1, maxxi=2.0,
               extramodels=False, rust=False, mlcp=True, debug=False):
    """
    Passed data from the Generalized Pareto Distribution, generate random samples from the GPD with calibrating prior.

    Parameters
    ----------
    n : int
        Number of samples to generate.
    x : array_like
        Input data array.
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    mlcp : bool, optional
        Whether to use ML and CP deviates (default is True).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML deviates, CP deviates, RU deviates, and method info.
    """
    
    if ics is None:
        ics = np.array([0, 0])
    
    x = cp_utils.to_array(x)
    
    # stopifnot(is.finite(n),!is.na(n),is.finite(x),!is.na(x),length(ics)==2,!x<0)
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert len(ics) == 2, "ics must have length 2"
    assert not np.any(x < 0), "x must be non-negative"
    
    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    cp_deviates = "rust not selected"
    
    if mlcp:
        q = ppf(x, np.random.uniform(0, 1, n), kloc=kloc, ics=ics, extramodels=extramodels)
        ml_params = q['ml_params']
        ml_deviates = q['ml_quantiles']
        ru_deviates = q['ru_quantiles']
        cp_deviates = q['cp_quantiles']
    
    if rust:
        th = tgpd_k1_cp(n, x)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            c = -th[i, 1]  # Convert xi to scipy parameterization
            ru_deviates[i] = stats.genpareto.rvs(c, loc=kloc, scale=th[i, 0], size=1)[0]
    
    op = {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.crhpflat_dmgs_cpmethod()
    }
    
    return op


def pdf(x, y=None, kloc=0, ics=np.array([0, 0]),
               minxi=-1, maxxi=2.0,
               rust=False, nrust=1000):
    """
    Passed data from the Generalized Pareto Distribution, compute the density function for the GPD with calibrating prior.

    Parameters
    ----------
    x : array_like
        Input data array.
    y : array_like, optional
        Points at which to evaluate the density (default is x).
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML PDF, RU PDF, and method info.
    """

    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y)), "y must be finite and not NA"
    assert len(ics) == 2, "ics must have length 2"
    assert not np.any(x < 0), "x must be non-negative"
    assert not np.any(y < 0), "y must be non-negative"
    
    ics = cp_gpd_b.gpd_k1_setics(x, ics)
    opt1 = optimize.minimize(lambda params: -cp_gpd_b.gpd_k1_loglik(params, x, kloc), 
                           ics, method='BFGS')
    v1hat = opt1.x[0]
    v2hat = opt1.x[1]
    
    if v2hat <= (-1):
        revert2ml = True
    else:
        revert2ml = False
    
    ml_params = np.array([v1hat, v2hat])
    # gpd_k1_checkmle(ml_params,kloc,minxi,maxxi)
    
    dd = cp_gpd_b.dgpdsub(x=x, y=y, ics=ics, kloc=kloc,
                        minxi=minxi, maxxi=maxxi)
    ru_pdf = "rust not selected"
    
    if rust and not revert2ml:
        th = tgpd_k1_cp(nrust, x)['theta_samples']
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            c = -th[ir, 1]  # Convert xi to scipy parameterization
            ru_pdf = ru_pdf + stats.genpareto.pdf(y, c, loc=kloc, scale=th[ir, 0])
        ru_pdf = ru_pdf / nrust
    else:
        ru_pdf = dd['ml_pdf']
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_pdf': dd['ml_pdf'],
        'revert2ml': revert2ml,
        'cp_pdf': ru_pdf,
        'cp_method': cp_utils.nopdfcdfmsg()
    }
    
    return op


def cdf(x, y=None, kloc=0, ics = np.array([0, 0]), customprior=0,
               minxi=-1, maxxi=2.0, extramodels=False,
               rust=False, nrust=1000, debug=False):
    """
    Passed data from the Generalized Pareto Distribution, compute the cumulative distribution function for the GPD with calibrating prior.

    Parameters
    ----------
    x : array_like
        Input data array.
    y : array_like, optional
        Points at which to evaluate the CDF (default is x).
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML CDF, RU CDF, and method info.
    """
    
    if y is None:
        y = x
    
    x = cp_utils.to_array(x)
    y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y)), "y must be finite and not NA"
    assert len(ics) == 2, "ics must have length 2"
    assert not np.any(x < 0), "x must be non-negative"
    assert not np.any(y < 0), "y must be non-negative"
    
    ics = cp_gpd_b.gpd_k1_setics(x, ics)
    opt1 = optimize.minimize(lambda params: -cp_gpd_b.gpd_k1_loglik(params, x, kloc), 
                           ics, method='BFGS')
    v1hat = opt1.x[0]
    v2hat = opt1.x[1]
    
    if v2hat <= (-1):
        revert2ml = True
    else:
        revert2ml = False
    
    ml_params = np.array([v1hat, v2hat])
    # gpd_k1_checkmle(ml_params,kloc,minxi,maxxi)
    
    dd = cp_gpd_b.dgpdsub(x=x, y=y, ics=ics, kloc=kloc,
                        minxi=minxi, maxxi=maxxi)
    ru_cdf = "rust not selected"
    
    if rust and not revert2ml:
        th = tgpd_k1_cp(nrust, x)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            c = -th[ir, 1]  # Convert xi to scipy parameterization
            ru_cdf = ru_cdf + stats.genpareto.cdf(y, c, loc=kloc, scale=th[ir, 0])
        ru_cdf = ru_cdf / nrust
    else:
        ru_cdf = dd['ml_cdf']
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_cdf': dd['ml_cdf'],
        'revert2ml': revert2ml,
        'cp_cdf': ru_cdf,
        'cp_method': cp_utils.nopdfcdfmsg()
    }
    
    return op


def tgpd_k1_cp(n, x, kloc=0, ics=None, extramodels=False, debug=False):
    """
    Not implemented: Theta sampling for the GPD with calibrating prior.

    Parameters
    ----------
    n : int
        Number of theta samples to generate.
    x : array_like
        Input data array.
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing theta samples.
    """
        
    raise Exception('tgpd_k1_cp (and rust) is not yet implemented in fitdistcp; please use the dmgs method.')
    
    if ics is None:
        ics = np.array([0, 0])
    
    x = cp_utils.to_array(x)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert len(ics) == 2, "ics must have length 2"
    assert not np.any(x < 0), "x must be non-negative"
    
    ics = cp_gpd_b.gpd_k1_setics(x, ics)
    t = cp_utils.ru(cp_gpd_b.gpd_k1_logf, x=x, kloc=kloc, n=n, d=2, init=ics)
    
    return {'theta_samples': t['sim_vals']}


import scipy
import sympy
import numpy as np
import autodiff

x, v1, v2 = sympy.symbols('x v1 v2')

f = (1/v2)*sympy.exp(-( ((x-v1)/v2) + sympy.exp(-((x-v1)/v2))))
p = sympy.exp(-(sympy.exp(-((x-v1)/v2))))
logf = -sympy.log(v2)-(x-v1)/v2-sympy.exp(-(x-v1)/v2)
                                 
with open('output_gumbel.py', 'w') as file:
    file.write('import numpy as np\n')
    file.write(autodiff.diff_to_code(f, [v1, v2], 'gumbel_fd', 1))
    file.write(autodiff.diff_to_code(f, [v1, v2], 'gumbel_fdd', 2))
    file.write(autodiff.diff_to_code(p, [v1, v2], 'gumbel_pd', 1))
    file.write(autodiff.diff_to_code(p, [v1, v2], 'gumbel_pdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'gumbel_logfdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'gumbel_logfddd', 3))

file.close()


import numpy as np
from scipy.stats import norm

def norm_fd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = -np.sqrt(2)*(2*v1 - 2*x)*np.exp(-(-v1 + x)**2/(2*v2**2))/(4*np.sqrt(np.pi)*v2**3) 
		result[1,i] = -np.sqrt(2)*np.exp(-(-v1 + x)**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**2) + np.sqrt(2)*(-v1 + x)**2*np.exp(-(-v1 + x)**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**4) 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def norm_fdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -np.sqrt(2)*(1 - (v1 - x)**2/v2**2)*np.exp(-(v1 - x)**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**3)
		result[0,1,i] = np.sqrt(2)*(3 - (v1 - x)**2/v2**2)*(v1 - x)*np.exp(-(v1 - x)**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**4)
		result[1,0,i] = np.sqrt(2)*(3 - (v1 - x)**2/v2**2)*(v1 - x)*np.exp(-(v1 - x)**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**4)
		result[1,1,i] = np.sqrt(2)*(1 - (3 - (v1 - x)**2/v2**2)*(v1 - x)**2/(2*v2**2) - (v1 - x)**2/v2**2)*np.exp(-(v1 - x)**2/(2*v2**2))/(np.sqrt(np.pi)*v2**3)
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 


def norm_pd(x, v1, v2):
    e1 = x - v1
    e3 = norm.pdf(e1/v2, 0, 1)
    result = np.column_stack([-(e3/v2), -(e3 * e1/v2**2)])
    return result


def norm_pdd(x, v1, v2):
    e1 = x - v1
    e2 = v2**2
    e4 = norm.pdf(e1/v2, 0, 1)
    e6 = e1**2/e2
    e7 = -((e6 - 1) * e4/e2)
    e8 = v2**3
    
    v1_v1 = -(e4 * e1/e8)
    v1_v2 = e7
    v2_v1 = e7
    v2_v2 = -((e6 - 2) * e4 * e1/e8)
    
    result = np.array([[v1_v1, v1_v2], 
                       [v2_v1, v2_v2]])
    return result


def norm_logfdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -1/v2**2
		result[0,1,i] = 2*(v1 - x)/v2**3
		result[1,0,i] = 2*(v1 - x)/v2**3
		result[1,1,i] = (1 - 3*(v1 - x)**2/v2**2)/v2**2
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def norm_logfddd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, dim, dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,0,i] = 0 
		result[0,0,1,i] = 2/v2**3 
		result[0,1,0,i] = 2/v2**3 
		result[0,1,1,i] = -6*(v1 - x)/v2**4 
		result[1,0,0,i] = 2/v2**3 
		result[1,0,1,i] = -6*(v1 - x)/v2**4 
		result[1,1,0,i] = -6*(v1 - x)/v2**4 
		result[1,1,1,i] = 2*(-1 + 6*(v1 - x)**2/v2**2)/v2**3 
	if len(x_all) == 1:
		result = result[:,:,:,0]
	return result 



import numpy as np


def format_data_a(actual_p_ml, actual_p_cp, p):
    x = []
    y_ml = []
    y_cp = []
    for i in range(len(p)):
        x.append(p[len(p)-1-i])
        y_ml.append(actual_p_ml[len(p)-1-i])
        y_cp.append(actual_p_cp[len(p)-1-i])
    x_range = (x[0], x[-1])
    y_range = (y_ml[0], y_ml[-1])
    return {'xlabel': 'Nominal Probability',  'ylabel': 'PCP', 'x':x, 'y_ml':y_ml, 'y_cp':y_cp, 
            'title': '(a) PCP vs NP', 'benchmark':{'x': x_range, 'y': x_range}, 'limits': {'x': x_range, 'y': y_range}}


def format_data_b(actual_p_ml, actual_p_cp, p):
    x = []
    y_ml = []
    y_cp = []
    for p_val in 0.0001 * np.asarray(range(1000,-1,-1)):
        x.append(p_val)
        y_ml.append(actual_p_ml[int(p_val*10000)])
        y_cp.append(actual_p_cp[int(p_val*10000)])
    x_range = (x[0], x[-1])
    y_range = x_range
    return {'xlabel': 'Nominal Probability',  'ylabel': 'PCP', 'x':x, 'y_ml':y_ml, 'y_cp':y_cp, 
            'title': '(b) PCP vs NP (tail)', 'benchmark':{'x': x_range, 'y': x_range}, 'limits': {'x': x_range, 'y': y_range}}


def format_data_d(actual_p_ml, actual_p_cp, p):
    x = []
    y_ml = []
    y_cp = []
    cut_off_index = 200 # since graph peaks sharply, can ignore first (p smallest) n points
    for i in range(len(p) - cut_off_index):
        i_rev = len(p) - 1 - i
        x.append(p[i_rev])
        y_ml.append(actual_p_ml[i_rev]/p[i_rev])
        y_cp.append(actual_p_cp[i_rev]/p[i_rev])
    x_range = (x[0], x[-1])
    y_range = (y_ml[0], y_ml[-1])
    return {'xlabel': 'Nominal Probability',  'ylabel': 'Prob Ratio', 'x':x, 'y_ml':y_ml, 'y_cp':y_cp, 
            'title': '(d) PCP/NP vs NP', 'benchmark':{'x': x_range, 'y': (1,1)}, 'limits': {'x': x_range, 'y': y_range}}


def format_data_h(actual_p_ml, actual_p_cp, p):
    x = []
    y_ml = []
    y_cp = []
    cut_off_index = 30
    for i in range(len(p) - cut_off_index):
        i_rev = len(p) - 1 - i
        x.append(1/p[i_rev])
        y_ml.append(1/actual_p_ml[i_rev])
        y_cp.append(1/actual_p_cp[i_rev])
    x_range = (x[0], x[-1])
    return {'xlabel': '1/Nominal Probability',  'ylabel': '1/PCP', 'x':x, 'y_ml':y_ml, 'y_cp':y_cp, 
            'title': '(h) 1/PCP vs 1/NP', 'benchmark':{'x': x_range, 'y': x_range}, 'limits': {'x': x_range, 'y': x_range}}


def format_data_i(actual_p_ml, actual_p_cp, p):
    x = []
    y_ml = []
    y_cp = []
    cut_off_index = 40  # cut off both the highest and lowest p values
    for i in range(len(p) - 2*cut_off_index):
        i_rev = len(p)-1-i-cut_off_index
        np = p[i_rev]
        pcp_ml = actual_p_ml[i_rev]
        pcp_cp = actual_p_cp[i_rev]
        x.append(1/np - 1/(1-np))
        y_ml.append(1/pcp_ml - 1/(1-pcp_ml))
        y_cp.append(1/pcp_cp - 1/(1-pcp_cp))
    x_range = (x[0], x[-1])
    return {'xlabel': '1/NP-1/(1-NP)',  'ylabel': '1/PCP-1/(1-PCP)', 'x':x, 'y_ml':y_ml, 'y_cp': y_cp,
            'title': '(h) PCP vs NP', 'benchmark':{'x': x_range, 'y': x_range}, 'limits': {'x': x_range, 'y': x_range}}





from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="fitdistcp",
    version="0.0.0",
    author="Lynne Jewson",
    author_email="lynne.jewson@gmail.com",
    description="A Python package for making reliable predictions using calibrating priors.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lynnejewson/fitdistcp",
    packages=find_packages(),
    install_requires=[
        'numpy',
        'scipy',
        'scikit-learn',
        'matplotlib'
    ]
)


import numpy as np
from scipy.stats import gumbel_r

def gumbel_fd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = (1/v2 - np.exp(-(-v1 + x)/v2)/v2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2 
		result[1,i] = ((-v1 + x)/v2**2 - (-v1 + x)*np.exp(-(-v1 + x)/v2)/v2**2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2 - np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**2 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def gumbel_fdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = ((np.exp((v1 - x)/v2) - 1)**2 - np.exp((v1 - x)/v2))*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**3
		result[0,1,i] = (2*np.exp((v1 - x)/v2) - 2 - (v1 - x)*(np.exp((v1 - x)/v2) - 1)**2/v2 + (v1 - x)*np.exp((v1 - x)/v2)/v2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**3
		result[1,0,i] = (2*np.exp((v1 - x)/v2) - 2 - (v1 - x)*(np.exp((v1 - x)/v2) - 1)**2/v2 + (v1 - x)*np.exp((v1 - x)/v2)/v2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**3
		result[1,1,i] = (2 - 2*(v1 - x)*(np.exp((v1 - x)/v2) - 1)/v2 - (v1 - x)*(2*np.exp((v1 - x)/v2) - 2 - (v1 - x)*(np.exp((v1 - x)/v2) - 1)**2/v2 + (v1 - x)*np.exp((v1 - x)/v2)/v2)/v2)*np.exp(-np.exp(-(-v1 + x)/v2) - (-v1 + x)/v2)/v2**3
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def gumbel_pd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = -np.exp(-(-v1 + x)/v2)*np.exp(-np.exp(-(-v1 + x)/v2))/v2 
		result[1,i] = -(-v1 + x)*np.exp(-(-v1 + x)/v2)*np.exp(-np.exp(-(-v1 + x)/v2))/v2**2 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def gumbel_pdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = (np.exp((v1 - x)/v2) - 1)*np.exp((v1 - x)/v2)*np.exp(-np.exp((v1 - x)/v2))/v2**2
		result[0,1,i] = (1 - (v1 - x)*np.exp((v1 - x)/v2)/v2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)*np.exp(-np.exp((v1 - x)/v2))/v2**2
		result[1,0,i] = (1 - (v1 - x)*np.exp((v1 - x)/v2)/v2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)*np.exp(-np.exp((v1 - x)/v2))/v2**2
		result[1,1,i] = (v1 - x)*(-2 + (v1 - x)*np.exp((v1 - x)/v2)/v2 - (v1 - x)/v2)*np.exp((v1 - x)/v2)*np.exp(-np.exp((v1 - x)/v2))/v2**3
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def gumbel_logfdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -np.exp((v1 - x)/v2)/v2**2
		result[0,1,i] = (np.exp((v1 - x)/v2) - 1 + (v1 - x)*np.exp((v1 - x)/v2)/v2)/v2**2
		result[1,0,i] = (np.exp((v1 - x)/v2) - 1 + (v1 - x)*np.exp((v1 - x)/v2)/v2)/v2**2
		result[1,1,i] = (1 - 2*(v1 - x)*np.exp((v1 - x)/v2)/v2 + 2*(v1 - x)/v2 - (v1 - x)**2*np.exp((v1 - x)/v2)/v2**2)/v2**2
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def gumbel_logfddd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, dim, dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,0,i] = -np.exp((v1 - x)/v2)/v2**3 
		result[0,0,1,i] = (2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)/v2**3 
		result[0,1,0,i] = (2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)/v2**3 
		result[0,1,1,i] = (-2*np.exp((v1 - x)/v2) + 2 - 4*(v1 - x)*np.exp((v1 - x)/v2)/v2 - (v1 - x)**2*np.exp((v1 - x)/v2)/v2**2)/v2**3 
		result[1,0,0,i] = (2 + (v1 - x)/v2)*np.exp((v1 - x)/v2)/v2**3 
		result[1,0,1,i] = (-2*np.exp((v1 - x)/v2) + 2 - 4*(v1 - x)*np.exp((v1 - x)/v2)/v2 - (v1 - x)**2*np.exp((v1 - x)/v2)/v2**2)/v2**3 
		result[1,1,0,i] = (-2*np.exp((v1 - x)/v2) + 2 - 4*(v1 - x)*np.exp((v1 - x)/v2)/v2 - (v1 - x)**2*np.exp((v1 - x)/v2)/v2**2)/v2**3 
		result[1,1,1,i] = (-2 + 6*(v1 - x)*np.exp((v1 - x)/v2)/v2 - 6*(v1 - x)/v2 + 6*(v1 - x)**2*np.exp((v1 - x)/v2)/v2**2 + (v1 - x)**3*np.exp((v1 - x)/v2)/v2**3)/v2**3 
	if len(x_all) == 1:
		result = result[:,:,:,0]
	return result 


############################################################
#' The first derivative of the density
def gumbel_f1fa(x,v1,v2):
	f1=gumbel_fd(x,v1,v2)
	return f1

############################################################
#' The second derivative of the density
def gumbel_f2fa(x,v1,v2):
	f2=gumbel_fdd(x,v1,v2)
	return f2

############################################################
#' The first derivative of the cdf
def gumbel_p1fa(x,v1,v2):
	p1=gumbel_pd(x,v1,v2)
	return p1

############################################################
#' The second derivative of the cdf
def gumbel_p2fa(x,v1,v2):
	p2=gumbel_pdd(x,v1,v2)
	return p2

############################################################
#' Minus the first derivative of the cdf, at alpha
def gumbel_mu1fa(alpha,v1,v2):
	x=gumbel_r.ppf((1-alpha),loc=v1,scale=v2)
	mu1=-gumbel_pd(x,v1,v2)
	return mu1

############################################################
#' Minus the second derivative of the cdf, at alpha
def gumbel_mu2fa(alpha,v1,v2):
	x=gumbel_r.ppf((1-alpha),loc=v1,scale=v2)
	mu2=-gumbel_pdd(x,v1,v2)
	return mu2

############################################################
#' The second derivative of the normalized log-likelihood
def gumbel_ldda(x,v1,v2):
	nx=len(x)
	temp1=gumbel_logfdd(x,v1,v2)
	ldd=np.sum(temp1, axis=-1) / nx
	return ldd

############################################################
#' The third derivative of the normalized log-likelihood
def gumbel_lddda(x,v1,v2):
	nx=len(x)
	temp1=gumbel_logfddd(x,v1,v2)
	lddd=np.sum(temp1, axis=-1) / nx
	return lddd



import numpy as np
import scipy.stats as stats
import scipy.optimize as optimize
import warnings
import sys

from . import utils as cp_utils
from . import genpareto_derivs as cp_gpd_k1_c


def rgpd_k1_minmax(nx, kloc, sigma, xi, minxi=-0.45, maxxi=0.45):
    """rgpd for gpd_k1 but with maxlik xi within bounds
    
    Returns:
        Vector
    """
    xihat = -9999
    while (xihat < minxi) or (xihat > maxxi):  # 0.46 also works...0.47 doesn't
        xx = stats.genpareto.rvs(c=xi, loc=kloc, scale=sigma, size=nx)
        ics = gpd_k1_setics(xx, np.array([0, 0]))
        opt1 = optimize.minimize(lambda params: -gpd_k1_loglik(params, xx, kloc), 
                               ics, method='BFGS')
        xihat = opt1.x[1]
    return xx

def gpd_k1_waic(waicscores, x, v1hat, v2hat, kloc, lddi, lddd, lambdad):
    """Waic
    
    Returns:
        Dictionary with waic1 and waic2
    """
    if waicscores:
        f1f = cp_gpd_k1_c.gpd_k1_f1fa(x, v1hat, v2hat, kloc)
        f2f = cp_gpd_k1_c.gpd_k1_f2fa(x, v1hat, v2hat, kloc)
        
        fhatx = stats.genpareto.pdf(x, c=v2hat, loc=kloc, scale=v1hat)
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim=2)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}


def gpd_k1_logf(params, x, kloc):
    """Logf for RUST"""
    sc = max(params[0], sys.float_info.epsilon)
    sh = params[1]
    logf = np.sum(stats.genpareto.logpdf(x, c=sh, loc=kloc, scale=sc)) - np.log(sc)
    return logf


def gpd_k1_setics(x, ics):
    """Set initial conditions"""
    if (ics[0] == 0) and (ics[1] == 0):
        ics_new = [np.std(x, ddof=1), 0]
    else:
        ics_new = ics
    return ics_new


def gpd_k1_loglik(vv, x, kloc):
    """log-likelihood function"""
    n = len(x)
    loglik = np.sum(stats.genpareto.logpdf(
        x, c=vv[1], loc=kloc, scale=max(vv[0], sys.float_info.epsilon)))
    return loglik


def gpd_k1_checkmle(ml_params, kloc, minxi=-1, maxxi=2):
    """Check MLE"""
    # currently not used, because instead I revert2ml
    v1hat = ml_params[0]
    v2hat = ml_params[1]
    if np.isnan(v1hat):
        raise Exception("v1hat is NaN")
    if np.isnan(v2hat):
        raise Exception("v2hat is NaN")
    
    # min xi
    # minxi=0
    if v2hat < minxi:
        warnings.warn(f"\n***v2hat={v2hat}=> execution halted because maxlik shape parameter <{minxi}***\n")
        raise Exception(f"v2hat={v2hat} < minxi={minxi}")
    
    # max xi
    if v2hat > maxxi:
        warnings.warn(f"\n***v2hat={v2hat}=> execution halted because maxlik shape parameter >{maxxi}***\n")
        raise Exception(f"v2hat={v2hat} > maxxi={maxxi}")
    
    # This max value is ad-hoc
    # If it's lowered to 1, then the ppm results for xi=0.6 go wrong, which I understand.
    # If it's increased to 100, then in about 1 in a billion cases, for nx=25,
    # the xi-hat value is very large and the code crashes because lddi can't be calculated.
    # I suspect there is more to understand about that latter part, but for now
    # this is a compromise.


def gpd_k1_means(means, ml_params, kloc=0):
    """Analytical Expressions for Predictive Means
    RHP mean based on the expectation of DMGS equation 2.1"""
    if means:
        # intro
        sigma = ml_params[0]
        xi = ml_params[1]
        
        # set up derivative arrays
        meand1 = np.zeros((2, 1))
        meand2 = np.zeros((2, 2, 1))
        
        # mle
        onemxi = 1 - xi
        onemxisq = onemxi * onemxi
        onemxicu = onemxi * onemxi * onemxi
        ml_mean = kloc + sigma / onemxi
        
        # calculate first derivative array for bayesian xi!=0 cases
        meand1[0, 0] = 1 / onemxi
        meand1[1, 0] = sigma / onemxisq
        
        # calculate second derivative array (only has 1 non-zero term!)
        meand2[0, 0, 0] = 0
        meand2[0, 1, 0] = 1 / onemxisq
        meand2[1, 1, 0] = -2 * sigma / onemxicu
        
        rh_flat_mean = np.inf
        
    else:
        ml_mean = "means not selected"
        rh_flat_mean = "means not selected"
    
    # return
    return {'ml_mean': ml_mean, 'rh_flat_mean': rh_flat_mean}


def dgpdsub(x, y, ics, kloc=0, minxi=None, maxxi=None):
    """Densities for 5 predictions """
    
    ics = gpd_k1_setics(x, ics)
    opt = optimize.minimize(lambda params: -gpd_k1_loglik(params, x, kloc), 
                          ics, method='BFGS')
    v1hat = opt.x[0]
    v2hat = min(maxxi, max(minxi, opt.x[1]))  # just reset in this case
    # v2hat = opt.x[1]
    ml_params = np.array([v1hat, v2hat])
    
    y = cp_utils.fixgpdrange(y, kloc, v1hat, v2hat)
    
    # mle
    ml_pdf = stats.genpareto.pdf(y, c=v2hat, loc=kloc, scale=v1hat)
    ml_cdf = stats.genpareto.cdf(y, c=v2hat, loc=kloc, scale=v1hat)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'ml_cdf': ml_cdf
    }


import sympy as sp
import numpy as np


def diff_to_functions(func, symbols, order):
    ''' Passed a symbolic function and symbols, differentiates and returns a matrix of derivatives. '''
    dim = len(symbols)
    if order == 1:
        result = np.empty(dim, dtype=object)
        for i in range(dim):
            result[i] = sp.diff(func, symbols[i])
    elif order == 2:
        result = np.empty((dim, dim), dtype=object)
        for i in range(dim):
            for j in range(dim):
                result[i, j] = sp.diff(func, symbols[i], symbols[j])
    elif order == 3:
        result = np.empty((dim, dim, dim), dtype=object)
        for i in range(dim):
            for j in range(dim):
                for k in range(dim):
                    result[i, j, k] = sp.diff(func, symbols[i], symbols[j], symbols[k])
    else:
        raise ValueError("Orders other than 1,2,3 are not supported.")
    return result


def numpify(text):
    '''Replaced exp and log in a string with np.exp and np.log'''
    return text.replace('exp(', 'np.exp(').replace('log(', 'np.log(')


def diff_to_code(func, symbols, func_title, order, xt=False):
    ''' Passed a symbolic function and symbols, differentiates once and returns a string of runnable Python code.'''
    func_deriv = diff_to_functions(func, symbols, order)       #vector of functions
    dim = len(symbols)

    string = '\n'
    diff_params = ''
    for i in range(len(symbols)):
        diff_params += ', v'+str(i+1)
    if not xt:
        string += ('def {}(x{}): \n'.format(func_title, diff_params))
    else:
        string += ('def {}(x, t{}): \n'.format(func_title, diff_params))
    string += '\t#Automatically generated code.\n'
    string += '\tdim={} \n'.format(dim)

    # vectorised, but works for non-vector input as well
    if xt:
        string += '\tt_is_vector = isinstance(t, list) or isinstance(t, np.ndarray)\n'
        string += '\tif t_is_vector:\n'
        string += '\t\tt_all = t\n'
    string += '\tif not (isinstance(x, list) or isinstance(x, np.ndarray)):\n'
    string += '\t\tx_all=[x]\n'
    string += '\telse:\n'
    string += '\t\tx_all=x\n'    

    if order == 1:
        string += '\tresult = np.zeros((dim, len(x_all)))\n'
        string += '\tfor i in range(len(x_all)):\n'
        string += '\t\tx = x_all[i]\n'
        if xt:
            string += '\t\tif t_is_vector:\n'
            string += '\t\t\tt = t_all[i]\n'
        for i in range(dim):
            string += '\t\tresult[{},i] = {} \n'.format(i, func_deriv[i])
        string += '\tif len(x_all) == 1:\n'
        string += '\t\tresult = result[:,0]\n'
        
    elif order == 2:
        string += '\tresult = np.zeros((dim,dim, len(x_all)))\n'
        string += '\tfor i in range(len(x_all)):\n'
        string += '\t\tx = x_all[i]\n'
        if xt:
            string += '\t\tif t_is_vector:\n'
            string += '\t\t\tt = t_all[i]\n'
        for i in range(dim):
            for j in range(dim):
                string += '\t\tresult[{},{},i] = {}\n'.format(i, j, func_deriv[i, j])
        string += '\tif len(x_all) == 1:\n'
        string += '\t\tresult = result[:,:,0]\n'

    elif order == 3:
        string += '\tresult = np.zeros((dim, dim, dim, len(x_all)))\n'
        string += '\tfor i in range(len(x_all)):\n'
        string += '\t\tx = x_all[i]\n'
        if xt:
            string += '\t\tif t_is_vector:\n'
            string += '\t\t\tt = t_all[i]\n'
        for i in range(dim):
            for j in range(dim):
                for k in range(dim):
                    string += '\t\tresult[{},{},{},i] = {} \n'.format(i, j, k, func_deriv[i, j, k])
        string += '\tif len(x_all) == 1:\n'
        string += '\t\tresult = result[:,:,:,0]\n'
    else:
        raise ValueError("Orders other than 1,2,3 are not supported.")
    
    string += '\treturn result \n'
    return numpify(string)


def diff_to_file(file, func, symbols, func_title, order, xt=False):
    ''' Passed an open file, appends Python code for the derivative function.
        symbols argument used to indicate which parameters to differentiate by (must be a list of sympy symbol objects).
        Options:
            xt=False:   the functions are of x only (vector or float)
            xt=True:    the functions are of x, t and works for 3 cases:
                1. x, t both floats
                2. x array and t a float (would be t0)
                3. x, t arrays of the same size.
        Returns a matrix with the last index corresponding to the x value.
    '''
    code = diff_to_code(func, symbols, func_title, order, xt=xt)
    file.write(code)


example = True
if example:
    # f is a symbolic function of x, v1, v2, v3, but only differentiate wrt v1, v2, v3
    x, t, v1, v2, v3 = sp.symbols('x t v1 v2 v3')
    f = x*v1**3+3*x*v1*v2**2*v3 + t*v2
    with open('autodiff_example_vectorised.py', 'w') as file:
        file.write('import numpy as np\n\n')
        diff_to_file(file, f, [v1, v2, v3], 'deriv_1', 1, xt=True)
        diff_to_file(file, f, [v1, v2, v3], 'deriv_2', 2, xt=True)
        diff_to_file(file, f, [v1, v2, v3], 'deriv_3', 3, xt=True)
    file.close()


import numpy as np
import math
import scipy.special as sp


def gamma_fd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = -x**(v1 - 1)*np.exp(-x/v2)*np.log(v2)/(v2**v1*math.gamma(v1)) + x**(v1 - 1)*np.exp(-x/v2)*np.log(x)/(v2**v1*math.gamma(v1)) - x**(v1 - 1)*np.exp(-x/v2)*sp.polygamma(0, v1)/(v2**v1*math.gamma(v1)) 
		result[1,i] = -v1*x**(v1 - 1)*np.exp(-x/v2)/(v2*v2**v1*math.gamma(v1)) + x*x**(v1 - 1)*np.exp(-x/v2)/(v2**2*v2**v1*math.gamma(v1)) 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def gamma_fdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = x**(v1 - 1)*(np.log(v2)**2 - 2*np.log(v2)*np.log(x) + 2*np.log(v2)*sp.polygamma(0, v1) + np.log(x)**2 - 2*np.log(x)*sp.polygamma(0, v1) + sp.polygamma(0, v1)**2 - sp.polygamma(1, v1))*np.exp(-x/v2)/(v2**v1*math.gamma(v1))
		result[0,1,i] = x**(v1 - 1)*(v1*np.log(v2) - v1*np.log(x) + v1*sp.polygamma(0, v1) - 1 - x*np.log(v2)/v2 + x*np.log(x)/v2 - x*sp.polygamma(0, v1)/v2)*np.exp(-x/v2)/(v2*v2**v1*math.gamma(v1))
		result[1,0,i] = x**(v1 - 1)*(v1*np.log(v2) - v1*np.log(x) + v1*sp.polygamma(0, v1) - 1 - x*np.log(v2)/v2 + x*np.log(x)/v2 - x*sp.polygamma(0, v1)/v2)*np.exp(-x/v2)/(v2*v2**v1*math.gamma(v1))
		result[1,1,i] = x**(v1 - 1)*(v1*(v1 + 1) - 2*v1*x/v2 - x*(2 - x/v2)/v2)*np.exp(-x/v2)/(v2**2*v2**v1*math.gamma(v1))
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def gamma_logfdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -sp.polygamma(1, v1)
		result[0,1,i] = -1/v2
		result[1,0,i] = -1/v2
		result[1,1,i] = (v1 - 2*x/v2)/v2**2
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def gamma_logfddd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, dim, dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,0,i] = -sp.polygamma(2, v1) 
		result[0,0,1,i] = 0 
		result[0,1,0,i] = 0 
		result[0,1,1,i] = v2**(-2) 
		result[1,0,0,i] = 0 
		result[1,0,1,i] = v2**(-2) 
		result[1,1,0,i] = v2**(-2) 
		result[1,1,1,i] = 2*(-v1 + 3*x/v2)/v2**3 
	if len(x_all) == 1:
		result = result[:,:,:,0]
	return result 

def gamma_f1fa(x, v1, v2):
    return gamma_fd(x, v1, v2)

def gamma_f2fa(x, v1, v2):
    return gamma_fdd(x, v1, v2)

def gamma_ldda(x, v1, v2):
    temp1 = gamma_logfdd(x, v1, v2)
    ldd = np.sum(temp1, axis=-1) / len(x)
    return ldd

def gamma_lddda(x, v1, v2):
    temp1 = gamma_logfddd(x, v1, v2)
    lddd = np.sum(temp1, axis=-1) / len(x)
    return lddd


import reltest
import os

file_name = "reltesttest genextreme 1.txt"
file_path = os.getcwd() + '\\reltest output\\' + file_name
#with open(file_path, 'w') as f:
#    reltest.reltest_genextreme(f)

# ...


with open(file_path, 'r') as f:
    reltest.plot(f)


import numpy as np
import scipy.stats
import scipy.special
import scipy.optimize
import sys

from . import utils as cp_utils
from . import evaluate_dmgs_equation as cp_dmgs
from . import weibull_derivs


def weibull_waic(waicscores, x, v1hat, v2hat, lddi, lddd, lambdad):
    """
    Waic for RUST
    
    Parameters
    ----------
    waicscores : bool
        Whether to calculate WAIC scores
    x : array_like
        Data values
    v1hat : float
        Shape parameter estimate
    v2hat : float
        Scale parameter estimate
    lddi : array_like
        Inverse of negative Hessian matrix
    lddd : array_like
        Third derivative array
    lambdad : array_like
        Lambda parameter for RHP
        
    Returns
    -------
    dict
        Dictionary containing waic1 and waic2 results
    """
    if waicscores:
        f1f = weibull_derivs.weibull_f1fa(x, v1hat, v2hat)
        f2f = weibull_derivs.weibull_f2fa(x, v1hat, v2hat)
        
        fhatx = scipy.stats.weibull_min.pdf(x, c=v1hat, scale=v2hat)
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim=2)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}

def weibull_logf(params, x):
    """
    Logf for RUST
    
    Parameters
    ----------
    params : array_like
        Parameters [shape, scale]
    x : array_like
        Data values
        
    Returns
    -------
    float
        Log-likelihood value with priors
    """
    sh = np.maximum(np.minimum(20, params[0]), np.sqrt(sys.float_info.epsilon))
    sc = np.maximum(params[1], np.sqrt(sys.float_info.epsilon))
    logf = np.sum(scipy.stats.weibull_min.logpdf(x, c=sh, scale=sc)) - np.log(sh) - np.log(sc)
    return logf

def weibull_loglik(vv, x):
    """
    log-likelihood function
    
    Parameters
    ----------
    vv : array_like
        Parameters [shape, scale]
    x : array_like
        Data values
        
    Returns
    -------
    float
        Log-likelihood value
    """
    n = len(x)
    shape_param = np.maximum(np.minimum(20, vv[0]), np.sqrt(sys.float_info.epsilon))
    scale_param = np.maximum(vv[1], np.sqrt(sys.float_info.epsilon))
    loglik = np.sum(scipy.stats.weibull_min.logpdf(x, c=shape_param, scale=scale_param))
    return loglik

def weibull_means(means, ml_params, lddi, lddd, lambdad_rhp, nx, dim=2):
    """
    MLE and RHP predictive means
    
    Parameters
    ----------
    means : bool
        Whether to calculate means
    ml_params : array_like
        ML parameter estimates [v1, v2]
    lddi : array_like
        Inverse of negative Hessian matrix
    lddd : array_like
        Third derivative array
    lambdad_rhp : array_like
        Lambda parameter for RHP
    nx : int
        Sample size
    dim : int, optional
        Dimension (default 2)
        
    Returns
    -------
    dict
        Dictionary containing ml_mean and rh_mean
    """
    # v1 is shape
    # v2 is scale
    if means:
        v1 = ml_params[0]
        v2 = ml_params[1]
        f1 = 1 + (1 / v1)
        iv12 = 1 / (v1 * v1)
        iv13 = 1 / (v1 * v1 * v1)
        
        # ml mean
        ml_mean = v2 * scipy.special.gamma(f1)
        
        # rhp mean
        meand1 = np.zeros((2, 1))
        meand1[0, 0] = -v2 * scipy.special.gamma(f1) * scipy.special.digamma(f1) * iv12
        meand1[1, 0] = scipy.special.gamma(f1)
        
        meand2 = np.zeros((2, 2, 1))
        meand2[0, 0, 0] = v2 * (scipy.special.gamma(f1) * scipy.special.polygamma(1, f1) * iv12 + 2 * scipy.special.gamma(f1) * scipy.special.digamma(f1) * iv13)
        meand2[0, 1, 0] = -scipy.special.gamma(f1) * scipy.special.digamma(f1) * iv12
        meand2[1, 0, 0] = meand2[0, 1, 0]
        meand2[1, 1, 0] = 0
        
        dmean = cp_dmgs.dmgs(lddi, lddd, meand1, lambdad_rhp, meand2, dim=2)
        rh_mean = ml_mean + dmean / nx
    else:
        ml_mean = "means not selected"
        rh_mean = "means not selected"
    
    return {'ml_mean': ml_mean, 'rh_mean': rh_mean}

def weibull_logscores(logscores, x):
    """
    Log scores for MLE and RHP predictions calculated using leave-one-out
    
    Parameters
    ----------
    logscores : bool
        Whether to calculate log scores
    x : array_like
        Data values
        
    Returns
    -------
    dict
        Dictionary containing ml_oos_logscore and rh_oos_logscore
    """
    if logscores:
        nx = len(x)
        ml_oos_logscore = 0
        rh_oos_logscore = 0
        
        for i in range(nx):
            # Create x1 by removing element at index i
            x1 = np.concatenate([x[:i], x[i+1:]])
            
            dd = dweibullsub(x1, x[i])
            
            ml_params = dd['ml_params']
            
            ml_pdf = dd['ml_pdf']
            ml_oos_logscore = ml_oos_logscore + np.log(ml_pdf)
            
            rh_pdf = dd['rh_pdf']
            rh_oos_logscore = rh_oos_logscore + np.log(np.maximum(rh_pdf, sys.float_info.epsilon))
    else:
        ml_oos_logscore = "logscores not selected"
        rh_oos_logscore = "logscores not selected"
    
    return {'ml_oos_logscore': ml_oos_logscore, 'rh_oos_logscore': rh_oos_logscore}

def dweibullsub(x, y):
    """
    Densities from MLE and RHP
    
    Parameters
    ----------
    x : array_like
        Training data
    y : float
        Prediction point
        
    Returns
    -------
    dict
        Dictionary containing ml_params, ml_pdf, rh_pdf, ml_cdf, rh_cdf
    """
    nx = len(x)
    
    # Optimization for MLE
    def neg_loglik(params):
        return -weibull_loglik(params, x)
    
    opt = scipy.optimize.minimize(neg_loglik, x0=np.array([1.0, 1.0]), method='BFGS')
    v1hat = opt.x[0]
    v2hat = opt.x[1]
    ml_params = np.array([v1hat, v2hat])
    
    # mle
    ml_pdf = scipy.stats.weibull_min.pdf(y, c=v1hat, scale=v2hat)
    ml_cdf = scipy.stats.weibull_min.cdf(y, c=v1hat, scale=v2hat)
    
    # rhp
    ldd = weibull_derivs.weibull_ldda(x, v1hat, v2hat)
    lddi = np.linalg.inv(ldd)
    lddd = weibull_derivs.weibull_lddda(x, v1hat, v2hat)
    
    f1 = weibull_derivs.weibull_f1fa(y, v1hat, v2hat)
    f2 = weibull_derivs.weibull_f2fa(y, v1hat, v2hat)
    
    p1 = weibull_derivs.weibull_p1fa(y, v1hat, v2hat)
    p2 = weibull_derivs.weibull_p2fa(y, v1hat, v2hat)
    
    lambdad_rhp = np.array([-1 / v1hat, -1 / v2hat])
    df1 = cp_dmgs.dmgs(lddi, lddd, f1, lambdad_rhp, f2, dim=2)
    dp1 = cp_dmgs.dmgs(lddi, lddd, p1, lambdad_rhp, p2, dim=2)
    rh_pdf = np.maximum(ml_pdf + df1 / nx, 0)
    rh_cdf = np.minimum(np.maximum(ml_cdf + dp1 / nx, 0), 1)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'rh_pdf': rh_pdf,
        'ml_cdf': ml_cdf,
        'rh_cdf': rh_cdf
    }
